# include <glib.h>
# include <stdio.h>
# include <netinet/in.h>
# include <sys/socket.h>
# include <signal.h>
# include <string.h>
# include <time.h>
# include "tpool.h"
# include "http.h"
# include "util.h"
# include "conf.h"
# include "peer.h"
# include "firewall.h"
# include "config.h"

void capture_peer ( http_request *h );
void splash_peer ( http_request *h );
void handle_request( http_request *h );

static GHashTable *peer_tab; 
G_LOCK_DEFINE(peer_tab);

/************* Permit and deny peers *************/

peer *find_peer ( const char *ip ) {
    peer *p; 
    
    p = g_hash_table_lookup( peer_tab, ip );
    if (p == NULL) {
	p = peer_new( ip );
	g_hash_table_insert( peer_tab, (gpointer) p->ip, p );
    }
    return p;
}

void accept_peer ( http_request *h ) {
    peer *p;
   
    G_LOCK(peer_tab);
    p  = find_peer( h->peer_ip );

    g_message( "Accepting peer %s", p->ip );

    peer_permit( nocat_conf, p );
    G_UNLOCK(peer_tab);

    http_add_header ( h, "Location", QUERY("redirect") );
    http_send_header( h, 302, "Moved" );
}

void remove_peer ( peer *p ) {
    g_message( "Removing peer %s", p->ip );
    peer_deny( nocat_conf, p );
}

gboolean check_peer_expire ( gchar *ip, peer *p, time_t *now ) {
    g_message( "Checking peer %s for expire: %lu sec. remain",
	ip, p->expire - *now );
    if (p->expire <= *now) {
	remove_peer( p );
	return TRUE;
    } else {
	return FALSE;
    }
} 

void check_peers( void *dummy ) {
    time_t now = time(NULL);
    g_message("Checking peers for expiration");
    G_LOCK(peer_tab);
    g_hash_table_foreach_remove( peer_tab, (GHRFunc)check_peer_expire, &now );
    G_UNLOCK(peer_tab);
}

gboolean run_peer_check( tpool_t pool ) {
    int r;

    // g_message("Adding peer check to work pool");
    r = tpool_add_work( pool, check_peers, NULL );
    if (r != 0) {
	g_warning( "Can't create thread to check peer table! Queue full?" );
	check_peers( NULL );
    }
    return TRUE;
}

/************* Connection handlers ************/

void handle_read( http_request *h ) {
    g_message( "thread entering handle_read" );
    http_request_read( h );

    if (http_request_ok(h))
	handle_request(h);

    g_io_channel_close( h->sock );
    g_io_channel_unref( h->sock );
    http_request_free( h );
    g_message( "thread exiting handle_read" );
}

gboolean handle_accept( GIOChannel *sock, GIOCondition cond, tpool_t pool ) {
    GIOChannel *conn;
    http_request *req;
    int fd, r;

    fd = accept( g_io_channel_unix_get_fd(sock), NULL, NULL );
    g_assert( fd != -1 );

    conn = g_io_channel_unix_new( fd );
    req  = http_request_new( conn );

    r = tpool_add_work( pool, (tpool_routine)handle_read, req );
    if (r != 0) {
	g_warning( "Can't create thread! Queue full?" );
	g_io_channel_unref( conn );
	http_request_free( req );
    }
    return TRUE;
}


/************* main ************/

static int caught_sigint = 0;

gboolean check_sigint ( GMainLoop *loop ) {
    if (caught_sigint) {
	g_message( "Caught SIGINT!" );
	g_main_quit( loop );
    }
    return TRUE;
}

void handle_sigint ( int sig ) {
    caught_sigint++;
}

int main (int argc, char **argv) {
    GMainLoop  *loop;
    GIOChannel *sock;
    tpool_t pool;

    /* Initialize signal handlers */
    signal(SIGPIPE, SIG_IGN);
    signal(SIGINT,  handle_sigint);

    /* read nocat.conf */
    read_conf_file( NC_CONF_PATH "/nocat.conf" );

    /* initialize the firewall */
    fw_init( nocat_conf );

    /* initialize the peer table */
    peer_tab = g_hash_new();

    /* initialize the listen socket */
    sock = http_bind_socket( 
	    CONF("GatewayAddr"), CONFd("GatewayPort"), CONFd("ListenQueue") );

    /* initialize the threading subsystem */
    g_thread_init(NULL);
    tpool_init( &pool, 5, 25, 0 );

    /* initialize the main loop and handlers */
    loop = g_main_new(FALSE);
    g_io_add_watch( sock, G_IO_IN,  (GIOFunc) handle_accept, pool );
    g_timeout_add( 5000, (GSourceFunc) run_peer_check, pool );
    g_timeout_add( 1000, (GSourceFunc) check_sigint, loop );
    
    /* Go! */
    g_message("starting main loop");
    g_main_run( loop );
    g_message("exiting main loop");
    return 0;
}
