/********************************************************
 * A modified version of the thread pool example from:
 *
 * "Using POSIX Threads: Programming with Pthreads"
 *     by Brad Nichols, Dick Buttlar, Jackie Farrell
 *     O'Reilly & Associates, Inc.
 *     ISBN 1-56592-115-1
 *
 * Copyright (C) 1996 O'Reilly & Associates, Inc.
 *
 * Used here under terms described in: 
 * http://www.oreilly.com/ask_tim/codepolicy_1101.html
 *
 ********************************************************
 * tpool.c -- 
 * 
 * Example thread pooling library
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <glib.h>
#include <pthread.h>
#include "tpool.h"
#include "util.h"

void *tpool_thread(void *);

void tpool_init(tpool_t   *tpoolp,
		int       num_worker_threads, 
		int       max_queue_size,
		int       do_not_block_when_full)
{
  int i, rtn;
  tpool_t tpool;
   
  /* allocate a pool data structure */ 
  if ((tpool = (tpool_t )g_new(struct tpool, 1)) == NULL)
    g_error("malloc: %m");

  /* initialize th fields */
  tpool->num_threads = num_worker_threads;
  tpool->max_queue_size = max_queue_size;
  tpool->do_not_block_when_full = do_not_block_when_full;
  if ((tpool->threads = 
       (pthread_t *)g_new(pthread_t, num_worker_threads))
      == NULL)
    g_error("malloc: %m");
  tpool->cur_queue_size = 0;
  tpool->queue_head = NULL; 
  tpool->queue_tail = NULL;
  tpool->queue_closed = 0;  
  tpool->shutdown = 0; 
  if ((rtn = pthread_mutex_init(&(tpool->queue_lock), NULL)) != 0)
    g_error("pthread_mutex_init %m");
  if ((rtn = pthread_cond_init(&(tpool->queue_not_empty), NULL)) != 0)
    g_error("pthread_cond_init %m");
  if ((rtn = pthread_cond_init(&(tpool->queue_not_full), NULL)) != 0)
    g_error("pthread_cond_init %m");
  if ((rtn = pthread_cond_init(&(tpool->queue_empty), NULL)) != 0)
    g_error("pthread_cond_init %m");

  /* create threads */
  for (i = 0; i != num_worker_threads; i++) {
    if ((rtn = pthread_create( &(tpool->threads[i]),
			      NULL,
			      tpool_thread,
			      (void *) tpool)) != 0)
      g_error("pthread_create %d",rtn);
  }

  *tpoolp = tpool;
}

int tpool_add_work(
		   tpool_t          tpool,
		   tpool_routine    routine,
		   void             *arg)
{
  int rtn;
  tpool_work_t *workp;

  if ((rtn = pthread_mutex_lock(&(tpool->queue_lock))) != 0)
    g_error("pthread_mutex_lock %d",rtn);

  /* no space and this caller doesn't want to wait */
  if ((tpool->cur_queue_size == tpool->max_queue_size) &&
      tpool->do_not_block_when_full) {
    if ((rtn = pthread_mutex_unlock(&(tpool->queue_lock))) != 0)
      g_error("pthread_mutex_unlock %d",rtn);

    return -1;
  }

  while( (tpool->cur_queue_size == tpool->max_queue_size) &&
	(!(tpool->shutdown || tpool->queue_closed))  ) {

    if ((rtn = pthread_cond_wait(&(tpool->queue_not_full),
				 &(tpool->queue_lock))) != 0)
      g_error("pthread_cond_wait %d",rtn);

  }

  /* the pool is in the process of being destroyed */
  if (tpool->shutdown || tpool->queue_closed) {
    if ((rtn = pthread_mutex_unlock(&(tpool->queue_lock))) != 0)
      g_error("pthread_mutex_unlock %d",rtn);
 
    return -1;
  }


  /* allocate work structure */
  if (( workp = g_new(tpool_work_t, 1)) == NULL)
    g_error("malloc: %m");
  workp->routine = routine;
  workp->arg = arg;
  workp->next = NULL;

  g_debug("adder: adding an item %p", workp->routine);

  if (tpool->cur_queue_size == 0) {
    tpool->queue_tail = tpool->queue_head = workp;

     g_debug("adder: queue == 0, waking all workers");

    if ((rtn = pthread_cond_broadcast(&(tpool->queue_not_empty))) != 0)
      g_error("pthread_cond_signal %d",rtn);;
  } else {
    tpool->queue_tail->next = workp;
    tpool->queue_tail = workp;
  }

  tpool->cur_queue_size++; 
  if ((rtn = pthread_mutex_unlock(&(tpool->queue_lock))) != 0)
    g_error("pthread_mutex_unlock %d",rtn);
  return 0;
}

int tpool_destroy(tpool_t          tpool,
		  int              finish)
{
  int          i,rtn;
  tpool_work_t *cur_nodep;
  

  if ((rtn = pthread_mutex_lock(&(tpool->queue_lock))) != 0)
    g_error("pthread_mutex_lock %d",rtn);

  /* Is a shutdown already in progress? */
  if (tpool->queue_closed || tpool->shutdown) {
    if ((rtn = pthread_mutex_unlock(&(tpool->queue_lock))) != 0)
      g_error("pthread_mutex_unlock %d",rtn);
    return 0;
  }

  tpool->queue_closed = 1;

  /* If the finish flag is set, wait for workers to 
     drain queue */ 
  if (finish == 1) {
    while (tpool->cur_queue_size != 0) {
      if ((rtn = pthread_cond_wait(&(tpool->queue_empty),
				   &(tpool->queue_lock))) != 0)
	g_error("pthread_cond_wait %d",rtn);
    }
  }

  tpool->shutdown = 1;

  if ((rtn = pthread_mutex_unlock(&(tpool->queue_lock))) != 0)
    g_error("pthread_mutex_unlock %d",rtn);


  /* Wake up any workers so they recheck shutdown flag */
  if ((rtn = pthread_cond_broadcast(&(tpool->queue_not_empty))) != 0)
    g_error("pthread_cond_broadcast %d",rtn);
  if ((rtn = pthread_cond_broadcast(&(tpool->queue_not_full))) != 0)
    g_error("pthread_cond_broadcast %d",rtn);


  /* Wait for workers to exit */
  for(i=0; i < tpool->num_threads; i++) {
    if ((rtn = pthread_join(tpool->threads[i],NULL)) != 0)
      g_error("pthread_join %d",rtn);
  }

  /* Now free pool structures */
  g_free(tpool->threads);
  while(tpool->queue_head != NULL) {
    cur_nodep = tpool->queue_head->next; 
    tpool->queue_head = tpool->queue_head->next;
    g_free(cur_nodep);
  }
  g_free(tpool); 
  return 0;
}

void *tpool_thread(void *arg)
{
  tpool_t tpool = (tpool_t)arg; 
  int rtn;
  tpool_work_t	*my_workp;
	
  for(;;) {
    /* Check queue for work */ 
    if ((rtn = pthread_mutex_lock(&(tpool->queue_lock))) != 0)
      g_error("pthread_mutex_lock %d",rtn);

    while ((tpool->cur_queue_size == 0) && (!tpool->shutdown)) {


      // g_debug("worker %d: I'm sleeping", (int)pthread_self());

      if ((rtn = pthread_cond_wait(&(tpool->queue_not_empty),
				   &(tpool->queue_lock))) != 0)
	g_error("pthread_cond_wait %d",rtn);

    }
    // g_debug("worker %d: I'm awake", (int)pthread_self());

    /* Has a shutdown started while i was sleeping? */
    if (tpool->shutdown == 1) {
      if ((rtn = pthread_mutex_unlock(&(tpool->queue_lock))) != 0)
	g_error("pthread_mutex_unlock %d",rtn);
      pthread_exit(NULL);
    }


    /* Get to work, dequeue the next item */ 
    my_workp = tpool->queue_head;
    tpool->cur_queue_size--;
    if (tpool->cur_queue_size == 0)
      tpool->queue_head = tpool->queue_tail = NULL;
    else
      tpool->queue_head = my_workp->next;
 
    g_debug("worker %d: dequeuing item %p",
	(int)pthread_self(), my_workp->routine);

    /* Handle waiting add_work threads */
    if ((!tpool->do_not_block_when_full) &&
	(tpool->cur_queue_size ==  (tpool->max_queue_size - 1))) 

      if ((rtn = pthread_cond_broadcast(&(tpool->queue_not_full))) != 0)
	g_error("pthread_cond_broadcast %d",rtn);

    /* Handle waiting destroyer threads */
    if (tpool->cur_queue_size == 0)

      if ((rtn = pthread_cond_signal(&(tpool->queue_empty))) != 0)
	g_error("pthread_cond_signal %d",rtn);

    if ((rtn = pthread_mutex_unlock(&(tpool->queue_lock))) != 0)
      g_error("pthread_mutex_unlock %d",rtn);
      
    /* Do this work item */
    (*(my_workp->routine))(my_workp->arg);
    g_free(my_workp);
  } 
  return(NULL);            
}
